#!/usr/bin/env python
from __future__ import print_function

from . import araalictl as api
import argparse
import difflib
import glob
import json
import os
import shutil
import sys
from . import utils
import yaml

class NoAliasDumper(yaml.Dumper):
    def ignore_aliases(self, data):
        return True

class Graph:
    def __init__(self, name):
        self.name = name

        self.node_dict = {}
        self.in_dict = {}
        self.out_dict = {}
        self.node_num = 0
        self.link_count = 0

    def add_link(self, client, server):
        client_key = yaml.dump(client)
        server_key = yaml.dump(server)
        if client_key not in self.node_dict:
            if not client.node.obj["name"]:
                client.node.obj["name"] = "node_%s" % self.node_num
            self.node_dict[client_key] = client
            self.node_num += 1
        
        if server_key not in self.node_dict:
            if not server.node.obj["name"]:
                server.node.obj["name"] = "node_%s" % self.node_num
            self.node_dict[server_key] = server
            self.node_num += 1
        
        self.out_dict.setdefault(self.node_dict[client_key].node.obj["name"],
            []).append(self.node_dict[server_key].node.obj["name"])
        
        self.in_dict.setdefault(self.node_dict[server_key].node.obj["name"],
            []).append(self.node_dict[client_key].node.obj["name"])

        self.link_count += 1
        
    def dump(self):
        basename = "%s/%s.yaml" % (utils.cfg.get("out_dir", utils.cfg["template_dir"]), self.name.replace("/", "_"))
        fname = basename

        files = glob.glob(fname+".*")
        nums = [int(a.rsplit(".", 1)[-1]) for a in files]
        if nums:
            num = 1+max(nums)
        else:
            num = 1

        fname = fname + ".%s" % num

        if os.path.isfile(basename): # yaml already exists
            os.rename(basename, fname) # backup current yaml
            if g_debug: print("backing up", basename, "to", fname)
        else:
            fname = None # we didnt need to backup

        with open(basename, "w") as f:
            yaml.dump(self, f, default_flow_style=False, default_style='', canonical=False, Dumper=NoAliasDumper)
        if g_debug: print("saving template", basename, "links:", self.link_count)

        if fname: # dont want to keep a backup if nothing changed
            with open(fname) as prev, open(basename) as curr:
                if next(difflib.unified_diff(prev.readlines(), curr.readlines()), None) == None:
                    # there is no difference, remove this one
                    os.remove(fname)
                    if g_debug: print("removing", fname, "same as", basename)

g_debug = False
def represent_graph(dumper, obj):
    value = []

    node_key = dumper.represent_data("identities")
    values = [a for a in obj.node_dict.values()]
    values.sort(key=lambda x: x.node.obj["name"])
    node_value = dumper.represent_data(values)
    value.append((node_key, node_value))

    if 0: # simple way to dump edges, to test the more optimized way below
        node_key = dumper.represent_data("edges")
        values = []
        for k,v in obj.out_dict.items():
            values.append(NameWithInOut(k, outlist=v))
        values.sort(key=lambda x: x.name)
        node_value = dumper.represent_data(values)
        value.append((node_key, node_value))

        return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

    ###
    # If it has more than one egress, we keep it, make peer ineligible
    #   if the peer is in third column, to keep things consistent. We dont
    #   want peer to sometiems be in out and sometimes be a node itself. We
    #   make it ineligible to become a node ever.
    #
    # If it has only one out, we can potentially make it "in" for some node
    #   if the peer is ineligible, we keep it
    #   if the peer has out (middle column), we remove it and keep the peer
    #     keep the peer, mark peer's peers as ineligible for in
    #   if the peer has only one in, we keep it (wont know in first pass)
    #   if the peer has multiple in, we keep the peer
    ###
    values = []
    ineligible = set() # this is a set of ineligible servers (peers)
    eligible = {}
    reevaluate = set()
    for k,v in obj.out_dict.items():
        if k in eligible: # we already took care of it
            if g_debug: print(k, "eligible", eligible)
            continue

        if len(v) > 1: # we keep it if it has more than one egress
            n = NameWithInOut(k, outlist=v)
            eligible[k] = n
            values.append(n)
            # since we consumed peer, we should always consume, even in other
            # scenarios/links - to be consistent
            for i in v:
                if i in ineligible: # it is already marked ineligible
                    continue
                if i not in obj.out_dict: # it is a pure server (3rd column)
                    if g_debug: print("for", k, "marking", i, "ineligible")
                    ineligible.add(i)
            continue

        # single peer case
        peer = v[0]
        if peer in ineligible:
            if g_debug: print("adding singleton", k, "because", peer, "ineligible")
            n = NameWithInOut(k, outlist=v)
            eligible[k] = n
            values.append(n)
            # peer is already marked ineligible
            continue

        # if peer is also a valid node with egress, we now have one ingress
        # coming into it, keep the peer
        if peer in obj.out_dict:
            # its not already added
            if peer not in eligible:
                if g_debug: print("processing singleton", k, "adding peer with egress", peer)
                n = NameWithInOut(peer, outlist=obj.out_dict[peer], inlist=[k])
                eligible[peer] = n
                values.append(n)
                for i in obj.out_dict[peer]:
                    if i not in obj.out_dict:
                        if g_debug: print("for", peer, "making", i, "ineligible")
                        ineligible.add(i)
                if peer in reevaluate:
                    reevaluate.remove(peer)
            else: # peer is already in, add me as ingress
                if g_debug: print("adding", k, "as additional ingress to existing", peer)
                eligible[peer].inlist.append(k)
        else: # peer has no egress, it might get marked ineligible
            reevaluate.add(k)

    # only single peer case need to be re-evaluated
    if g_debug: print("reevaluate", reevaluate)
    for k in reevaluate:
        peer = obj.out_dict[k][0]
        if peer in ineligible:
            values.append(NameWithInOut(k, outlist=[peer]))
        else: # peer is not ineligible
            if peer in eligible: # append to it
                eligible[peer].inlist.append(k)
            else: # adding peer for the first time
                n = NameWithInOut(peer, inlist=[k])
                eligible[peer] = n
                values.append(n)

    values.sort(key=lambda x: x.name)

    node_key = dumper.represent_data("authorizations")
    node_value = dumper.represent_data(values)
    value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

class NameWithInOut:
    def __init__(self, name, inlist=[], outlist=[]):
        self.name = name
        self.inlist = [a for a in inlist]
        self.outlist = [a for a in outlist]

def represent_name_with_in_out(dumper, obj):
    value = []

    node_key = dumper.represent_data("name")
    node_value = dumper.represent_data(obj.name)
    value.append((node_key, node_value))

    if obj.inlist:
        node_key = dumper.represent_data("in")
        obj.inlist.sort()
        node_value = dumper.represent_data(obj.inlist)
        value.append((node_key, node_value))

    if obj.outlist:
        node_key = dumper.represent_data("out")
        obj.outlist.sort()
        node_value = dumper.represent_data(obj.outlist)
        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

class NodeWithPushdown:
    def __init__(self, node, pushdown):
        self.node = node
        self.pushdown = pushdown

    def pushdown_dict(self):
        return dict([(k, self.node.obj[k]) for k in self.pushdown.obj])

def represent_node_with_pushdown(dumper, obj):
    value = []

    node_key = dumper.represent_data("node")
    node_value = dumper.represent_data(obj.node)
    value.append((node_key, node_value))

    node_key = dumper.represent_data("pushdown")
    node_value = dumper.represent_data(obj.pushdown)
    value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

class Pushdown:
    def __init__(self, obj):
        if type(obj) == dict:
            keys = [a for a in obj.keys()]
        else:
            keys = [a for a in obj]
        keys.sort()
        self.obj = keys

def represent_pushdown(dumper, obj):
    return dumper.represent_data(obj.obj)

class Node:
    def __init__(self, obj, name="__no_name__"):
        """Node is used by both node and pushdown, pushdown has no name"""
        self.obj = obj
        if name != "__no_name__":
            self.obj["name"] = name
        if "rename" in self.obj:
            self.obj["name"] = self.obj["rename"]
            del self.obj["rename"]

def represent_node(dumper, obj):
    def rank(key):
        key = key[0].value
        key_order = ["name", "dns_pattern", "subnet", "netmask", "dst_port",
                     "zone", "app", "pod", "container",
                     "binary_name", "parent_process", "process"]
        key_order = dict(zip(key_order, range(0, len(key_order))))
        return key_order[key]

    value = []
    for item_key, item_value in obj.obj.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    value.sort(key=lambda x: rank(x))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

def push(args):
    ans = input("are you sure? (yes/no)> ")
    if ans != "yes":
        print("This is a dangerous command, thanks for exercising caution!")
        return

    if args.verbose >= 1: print(args, ans)
    if args.verbose >= 1: api.g_debug = True

    basename = "%s/%s.yaml" % (utils.cfg["template_dir"], args.template)
    if not os.path.isfile(basename): # yaml already exists
        print(basename, "does not exist")
        return

    # read a sample template from araali
    obj = [a for a in api.API().get_templates(public=True, template="agent.k8s.araali_fw")[0]][0]
    # name, template, use, author, version
    if args.verbose >= 2: print(obj["name"], obj["use"], obj["author"], obj["version"])
    obj["name"] = args.template
    if args.public:
        obj["use"] = False
    else:
        obj["use"] = True
    obj["author"] = "Template As Code (Git)"
    obj["version"] = "v0.0.1"

    nodes = {}
    obj["template"] = []
    with open(basename) as f:
        template = yaml.load(f, Loader=yaml.SafeLoader)
        # load identities
        for node in template["identities"]:
            name = node["node"]["name"]
            nodes[name] = NodeWithPushdown(Node(node["node"]), Pushdown(node["pushdown"]))

        for node in template["authorizations"]:
            # name, in, out
            node_obj = nodes[node["name"]]
            if args.verbose >= 2: print(yaml.dump(node_obj))
            for label in node.get("in", []):
                peer_obj = nodes[label]
                o = {"link_filter": {}}
                o["link_filter"]["client"] = dict(peer_obj.node.obj)
                o["link_filter"]["server"] = dict(node_obj.node.obj)
                if node_obj.pushdown.obj or peer_obj.pushdown.obj:
                    o["selector_change"] = {}
                    if peer_obj.pushdown.obj:
                        o["selector_change"]["client"] = peer_obj.pushdown_dict()
                    if node_obj.pushdown.obj:
                        o["selector_change"]["server"] = node_obj.pushdown_dict()
                obj["template"].append(o)
            for label in node.get("out", []):
                peer_obj = nodes[label]
                o = {"link_filter": {}}
                o["link_filter"]["client"] = dict(node_obj.node.obj)
                o["link_filter"]["server"] = dict(peer_obj.node.obj)
                if node_obj.pushdown.obj or peer_obj.pushdown.obj:
                    o["selector_change"] = {}
                    if node_obj.pushdown.obj:
                        o["selector_change"]["client"] = node_obj.pushdown_dict()
                    if peer_obj.pushdown.obj:
                        o["selector_change"]["server"] = peer_obj.pushdown_dict()
                obj["template"].append(o)
 
    if args.verbose >= 1: print(obj)
    if args.tenant:
        tenant = args.tenant
    else:
        tenant = utils.cfg["tenant"]
    rc = api.API().update_template([obj], args.public, tenant)
    print(rc)

def alerts(args):
    if args.tenant:
        tenants = [{"name": args.tenant, "id": args.tenant}]
    elif utils.cfg.get("tenants", None):
        tenants = utils.cfg["tenants"]
        if not tenants:
            tenants = [{"name": utils.cfg["tenant"], "id": utils.cfg["tenant"]}]
        else:
            # we are doing a full run
            if not args.nopull:
                shutil.rmtree("%s/%s" % (args.progdir, ".alerts.template.py"), ignore_errors=True)
    else:
        tenants = [{"name": "account", "id": utils.cfg["tenant"]}]

    if args.nopull:
        if args.ago:
            import datetime
            def make_map(kv):
                k, v = kv.split("=")
                return k, int(v)
            ago = dict([make_map(a) for a in args.ago.split(",")])
            cutoff = datetime.datetime.now() - datetime.timedelta(**ago)
            for t in tenants:
                count = 0
                with open("%s/%s/%s.json" % (args.progdir, ".alerts.template.py", t["name"]), "r") as f:
                    for line in f:
                        line = json.loads(line)
                        # print(line.keys())
                        alert_t = datetime.datetime.fromtimestamp(int(line["timestamp"])/1000)
                        if alert_t > cutoff:
                            count += 1
                if count:
                    print("%-60s %s" % (t, count))
        else:
            print("ls .alerts.template.py/*.json | xargs wc -l")
        return

    if args.verbose >= 1: api.g_debug = True
    os.makedirs("%s/%s" % (args.progdir, ".alerts.template.py"), exist_ok=True)
    for t in tenants:
        with open("%s/%s/%s.json" % (args.progdir, ".alerts.template.py", t["name"]), "w") as f:
            count = 0
            skipped_count = 0
            token = None
            while True:
                alerts, _, _ = api.API().get_alerts(count=200000, token=token, tenant=t["id"])
                if not alerts:
                    break

                for obj in alerts:
                    if obj["alert_info"]["status"] == "CLOSE":
                        skipped_count += 1
                        continue
                    count += 1
                    json.dump(obj, f)
                    f.write("\n")

                token=obj.get("paging_token", None)
                if not token:
                    break

            print(t, "open=%s" % count, "closed=%s" % skipped_count)

def eval_drift(tenant):
    git_files = set([os.path.split(a)[1] for a in glob.glob('%s/*.yaml' % utils.cfg["template_dir"])])
    drift_files =set([os.path.split(a)[1] for a in glob.glob('%s/*.yaml' % utils.cfg["out_dir"])])

    not_in_git = list(drift_files - git_files)
    in_git = list(drift_files & git_files)
    
    not_in_git.sort()
    in_git.sort()
    
    print('\nEvaluating drift for:', tenant)
    if g_debug: print("git:", git_files, "drift:", drift_files)
    if not_in_git:
        print("*** Files not in git:")
        for f in not_in_git:
            print("\t", f)
    
    if tenant == "public":
        not_in_public = list(git_files - drift_files)
        not_in_public.sort()
        if not_in_public:
            print("*** Files not in public:")
            for f in not_in_public:
                print("\t", f)
            
    for fname in in_git:
        prevfname = utils.cfg["template_dir"] + "/" + fname
        currfname = utils.cfg["out_dir"] + "/" + fname
        with open(prevfname) as prev, open(currfname) as curr:
            if next(difflib.unified_diff(prev.readlines(), curr.readlines()), None) != None:
                print("***", prevfname, currfname, "have drifted")
            else:
                if g_debug:
                    print(prevfname, currfname, "have not drifted")


def drift(args):
    if not args.public:
        if args.tenant:
            tenants = [{"name": [a for a in utils.cfg["tenants"] if a["id"] == args.tenant][0]["name"], "id": args.tenant}]
        else:
            tenants = utils.cfg.get("tenants", None)
            if not tenants:
                tenants = [{"name": "account", "id": utils.cfg["tenant"]}]
            else:
                # full run
                if not args.nopull:
                    shutil.rmtree("%s/%s" % (args.progdir, ".drift.template.py"), ignore_errors=True)

    api.g_debug = False
    os.makedirs("%s/%s" % (args.progdir, ".drift.template.py"), exist_ok=True)

    if args.public:
        utils.cfg["out_dir"] = "%s/%s/public" % (args.progdir, ".drift.template.py")
        if not args.nopull:
            shutil.rmtree("%s/%s/public" % (args.progdir, ".drift.template.py"), ignore_errors=True)
            os.makedirs("%s/%s/public" % (args.progdir, ".drift.template.py"), exist_ok=True)
            args.tenant = None
            pull(args)
        eval_drift("public")
        return

    for t in tenants:
        utils.cfg["out_dir"] = "%s/%s/%s" % (args.progdir, ".drift.template.py", t["name"])
        if not args.nopull:
            shutil.rmtree("%s/%s/%s" % (args.progdir, ".drift.template.py", t["name"]), ignore_errors=True)
            os.makedirs("%s/%s/%s" % (args.progdir, ".drift.template.py", t["name"]), exist_ok=True)
            args.tenant = t["id"]
            pull(args)
        eval_drift("%s (%s)" % (t["name"], t["id"]))

def ls(args):
    if args.tenant:
        tenant = args.tenant
    else:
        tenant = utils.cfg["tenant"]
    for obj in api.API().get_templates(public=args.public, tenant=tenant,
                                       template=args.template)[0]:
        # keys: ['name', 'template', 'use', 'author', 'version']
        del obj["template"]
        print(obj)

def pull(args):
    if args.verbose >= 1: api.g_debug = True

    if args.tenant:
        tenant = args.tenant
    else:
        tenant = utils.cfg["tenant"]

    found = False
    for obj in api.API().get_templates(public=args.public, tenant=tenant,
                                       template=args.template)[0]:
        # get the names from local file and apply it to pulled template
        existing_nodes = {}

        basename = "%s/%s.yaml" % (utils.cfg["template_dir"], obj["name"])
        if os.path.isfile(basename): # yaml already exists
            with open(basename) as f:
                template = yaml.load(f, Loader=yaml.SafeLoader)

                # identities, authorizations
                for node in template["identities"]:
                    name = node["node"]["name"]
                    node["node"]["name"] = None
                    node = NodeWithPushdown(Node(node["node"]), Pushdown(node["pushdown"]))
                    key = yaml.dump(node)
                    existing_nodes[key] = name
 
        # keys: ['name', 'template', 'use', 'author', 'version']
        graph = Graph(obj["name"])
        for link in obj["template"]:
            #print(yaml.dump(link))

            link_filter = link["link_filter"]

            client = NodeWithPushdown(Node(link_filter["client"], None),
                        Pushdown(link.get("selector_change", {}).get("client", [])))
            server = NodeWithPushdown(Node(link_filter["server"], None),
                        Pushdown(link.get("selector_change", {}).get("server", [])))

            # fix up names based on existing nodes
            key = yaml.dump(client)
            if key in existing_nodes:
                client.node.obj["name"] = existing_nodes[key]
            else:
                print("not found", yaml.dump(client))

            key = yaml.dump(server)
            if key in existing_nodes:
                server.node.obj["name"] = existing_nodes[key]
            else:
                print("not found", yaml.dump(server))

            graph.add_link(client, server)

        graph.dump()
        found = True
    if not found:
        print("No template found")

def fmt(args):
    if args.verbose >= 1: print(args.template, args.dirname)
    graph = Graph(args.template)
    nodes = {}
    with open("%s/%s.yaml" % (args.dirname, args.template)) as f:
        template = yaml.load(f, Loader=yaml.SafeLoader)

        # identities, authorizations
        for node in template["identities"]:
            name = node["node"]["name"]
            nodes[name] = NodeWithPushdown(Node(node["node"]),
                                                   Pushdown(node["pushdown"]))
        for link in template["authorizations"]:
            for l in link.get("in", []):
                graph.add_link(nodes[l], nodes[link["name"]])
            for l in link.get("out", []):
                graph.add_link(nodes[link["name"]], nodes[l])
    graph.dump()

yaml.add_representer(Node, represent_node)
yaml.add_representer(Pushdown, represent_pushdown)
yaml.add_representer(NodeWithPushdown, represent_node_with_pushdown)
yaml.add_representer(Graph, represent_graph)
yaml.add_representer(NameWithInOut, represent_name_with_in_out)



from __future__ import print_function

import os
import shlex
import subprocess
import yaml

cfg_fname = os.environ['HOME']+"/.araalirc"

def read_config():
    if os.path.isfile(cfg_fname):
        with open(cfg_fname, "r") as f:
            cfg = yaml.load(f, Loader=yaml.SafeLoader)
    else:
        cfg = {}

    if not cfg.get("tenant", None): cfg["tenant"] = None
    if not cfg.get("template_dir", None): cfg["template_dir"] = "../templates"
    if not cfg.get("token", None): cfg["token"] = os.getenv('ARAALI_API_TOKEN')
    if not cfg.get("backend", None): cfg["backend"] = "prod"
    # env variable overrides on disk config
    if os.getenv("ARAALI_BACKEND"):
        cfg["backend"] = os.getenv("ARAALI_BACKEND")
    return cfg

cfg = read_config()

def config(tenant=None, tenants=None, template_dir=None, backend=None):
    global cfg

    def save(cfg):
        with open(cfg_fname, "w") as f:
            cfg2 = dict(cfg)
            del cfg2["token"]
            yaml.dump(cfg2, f)

    if tenant is not None:
        if not tenant: # passed as empty string
            tenant = None # store it as nil in yaml
        cfg["tenant"] = tenant
        save(cfg)
    if tenants is not None:
        cfg["tenants"] = [dict(zip(["name", "id"], a.split(":"))) for a in tenants.split(",")]
        save(cfg)
    for k in ["template_dir", "backend"]:
        if eval(k) is not None:
            cfg[k] = eval(k)
            save(cfg)
    print(yaml.dump(cfg))

def dump_table(objs):
    for idx, o in enumerate(objs):
        print("%s %s %s" % ("="*40, idx, "="*40))
        print(yaml.dump(o))

def make_map(kvstr):
    k, v = kvstr.split("=")
    return k, int(v)

def eprint(*args, **kwargs):
    args = [a.decode() for a in args]
    print(*args, file=sys.stderr, **kwargs)

def run_command(command, result=False, strip=True, in_text=None, debug=False, env=os.environ):
    def collect_output(ret, output):
        if output:
            if strip:
                output = output.strip()
            if result:
                ret.append(output.decode())
            else:
                eprint(output)

    if debug:
        print(command, result)
    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE,
                                stdin=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=env)
    ret = []
    if result:
        if in_text:
            outs, errs = process.communicate(input=in_text.encode())
        else:
            outs, errs = process.communicate()
        return process.poll(), outs+errs
    else:
        while True:
            rc = process.poll()
            err = process.stderr.readline()
            out = process.stdout.readline()
            if rc is not None and not out and not err:
                return rc, "\n".join(ret)
            collect_output(ret, out)
            collect_output(ret, err)



"""Used to execute modeule from the command line
    Fetches alerts and prints the count
"""
from __future__ import print_function

import argparse
import datetime
from dateutil import parser as du_parser
from dateutil import tz
import json
import os
import platform
import re
import shutil
import subprocess
import sys
import yaml

from araali import API
from . import api
from . import araalictl
from . import utils
from . import template as module_template

def config(args):
    if args.get_tenants:
        """
        Current:
            ./araalictl api -fetch-active-subtenants
            ./araalictl api -fetch-subtenants
            ./araalictl api -tenant=freemium -fetch-subtenants

        Desired:
            ./araalictl api -fetch-subtenants
            ./araalictl api -fetch-subtenants -active
        """
        if utils.cfg["backend"] != "prod": utils.config(backend="prod")
        tracking = [a for a in utils.cfg.get("tenants", [])]
        tracking_ids = [a["id"] for a in tracking]
        obj = araalictl.API().get_tenants()[0]["subtenantlist"]
        # ['tenantid', 'adminemail', 'activevmcount', 'activecontainercount', 'perimeteralertcount', 'homealertcount', 'lastsignedin']
        print("     %-22s %-42s %-7s %-10s %s" % ("tenant-id", "admin-email", "count", "tracked", "last-signed-in"))
        agent_count = 0
        for i, o in enumerate(obj):
            o["lastsignedin"] = datetime.datetime.fromtimestamp(o["lastsignedin"]/1000) if o["lastsignedin"] else ""
            print("%3s: t=%-20s e=%-40s c=%-5s tr=%-7s s=%s" % (i+1, o["tenantid"],
                o["adminemail"], o["activevmcount"], o["tenantid"] in tracking_ids,
                o["lastsignedin"] if o["lastsignedin"] else "nil"))
            agent_count += o["activevmcount"]
            tracking = [a for a in tracking if o["tenantid"] != a["id"]]
        print("Total agents: %s" % agent_count)
        if tracking:
            print("\n=== Tracking but not active ===")
            for i, t in enumerate(tracking):
                print("%3s: t=%-20s n=%s" % (i, t["id"], t["name"]))
        return
    return utils.config(args.tenant, args.tenants, args.template_dir, args.backend)

def alerts(args):
    alerts, page, status = API().get_alerts(args.count, args.ago, tenant=args.tenant)
    day_dict = {}
    local_zone = tz.tzlocal()
    if status == 0:
        print("Got %s alerts" % len(alerts))
        utils.dump_table(alerts)

        for a in alerts:
            if type(a["timestamp"]) == str:
                # '2022-07-15T04:45:18Z'
                dt = du_parser.parse(a["timestamp"]).astimezone(local_zone)
            else:
                dt = datetime.datetime.fromtimestamp(a["timestamp"]/1000)
            #dts = dt.strftime("%m/%d/%Y, %H:%M:%S")
            dts = dt.strftime("%Y/%m/%d")
            day_dict.setdefault(dts, []).append(a)

        keys = [a for a in day_dict.keys()]
        keys.sort()
        print("frequency count")
        for k in keys:
            print("%s %s" % (k, len(day_dict[k])))

def make_map(kv):
    k, v = kv.split("=")
    return k, v

def assets(args):
    if not args.nopull or not os.path.isfile("%s/%s/dump.json" % (args.progdir, ".assets.araali")):
        shutil.rmtree("%s/%s" % (args.progdir, ".assets.araali"), ignore_errors=True)
        os.makedirs("%s/%s" % (args.progdir, ".assets.araali"), exist_ok=True)
        with open("%s/%s/dump.json" % (args.progdir, ".assets.araali"), "w") as f:
            assets, status = API().get_assets(args.zone, args.app, args.ago, tenant=args.tenant)
            if status == 0:
                for a in assets:
                    json.dump(a, f)
                    f.write("\n")

                #print("Got %s assets" % len(assets))
                #utils.dump_table(assets)

    if args.filters:
        filters = dict([make_map(a) for a in args.filters.split(",")])
    else:
        filters = {}

    i = 0
    search_type = "asset"
    if set(["cve", "cve_sev"]) & set(filters):
        search_type = "cve"

    def cve_in_asset(line, filters):
        for v in line["vulnerabilities"]:
            for c in v["cve_id"]:
                if re.search(filters["cve"], c):
                    return True
        return False

    def match_filters(filters, line):
        def get_app_name(a):
            if type(a) == dict:
                return a["app_name"]
            return a

        for k,v in filters.items():
            if k == "app":
                if not re.search(":[^:]*" + v + "[^:]*:", ":"+ ":".join([get_app_name(a) for a in line["apps"]]) + ":"):
                    return False
            elif k not in line:
                return False
            else:
                if not re.search(v, line[k]):
                    return False
        return True

    cve_sev_map = {"NONE": 0, "LOW": 1, "MEDIUM": 2, "HIGH": 3, "CRITICAL": 4}
    with open("%s/%s/dump.json" % (args.progdir, ".assets.araali"), "r") as f:
        for line in f:
            line = json.loads(line)
            # 'name', 'ip_address', 'uuid', 'image', 'zone', 'apps', 'processes', 'state', 'asset_type', 'process_capabilities', 'ip_addresses', 'originaluuid', 'vulnerabilities', 'mode', 'os_name'
            if args.verbose > 1:
                print("="*40, i, "="*40)
                print("name", line["name"])
                print("ip", line["ip_address"])
                print("uuid", line["uuid"])
                print("orig_uuid", line["originaluuid"])
                print("image", line["image"])
                print("zone", line.get("zone", ""))
                print("apps", line.get("app", ""))
                #print("process", line.get("process", "")) # always empty
                print("state", line["state"])
                print("asset_type", line["asset_type"])
                #print("process_capabilities", line["process_capabilities"]) # always empty
                print("mode", line["mode"])
                print("os_name", line["os_name"])
                #print("ip_addresses", line["ip_addresses"]) # always empty
                # 'package_name', 'cve_id', 'severity'
                #print("cve", line["vulnerabilities"][0].keys()) # always empty

            line["cve_count"] = str(len(line.get("vulnerabilities", [])))
            if search_type == "cve":
                if "vulnerabilities" not in line or not line["vulnerabilities"]:
                    continue

                if "cve" in filters and cve_in_asset(line, filters):
                    print("="*40, i, "="*40)
                    i += 1
                    line["max_severity"] = max([cve_sev_map[a["severity"]] for a in line["vulnerabilities"]])
                    if not args.verbose:
                        del line["vulnerabilities"]
                    if "containerd" in line:
                        line["caps"] = len(line["containerd"]["capabilities"])
                        del line["containerd"]["capabilities"]
                    print(yaml.dump(line))
                elif "cve_sev" in filters and max([cve_sev_map[a["severity"]] for a in line["vulnerabilities"]]) == int(filters["cve_sev"]):
                    print("="*40, i, "="*40)
                    i += 1
                    line["max_severity"] = max([cve_sev_map[a["severity"]] for a in line["vulnerabilities"]])
                    del line["vulnerabilities"]
                    if "containerd" in line:
                        line["caps"] = len(line["containerd"]["capabilities"])
                        del line["containerd"]["capabilities"]
                    print(yaml.dump(line))

            else:
                if match_filters(filters, line):
                    print("="*40, i, "="*40)
                    i += 1
                    if "vulnerabilities" in line:
                        del line["vulnerabilities"]
                    if "containerd" in line:
                        line["caps"] = len(line["containerd"]["capabilities"])
                        #del line["containerd"]["capabilities"]
                    print(yaml.dump(line))


def links(args):
    links, status = API().get_links(args.zone, args.app, args.svc, args.ago, tenant=args.tenant)
    if status == 0:
        print("Got %s links" % len(links))
        utils.dump_table(links)

def search(args):
    def match_filters(filters, line):
        for k,v in filters.items():
            if k not in line:
                if args.verbose: print("***", k, "not in", line)
                return False
            else:
                if not re.search(v, line[k]):
                    if args.verbose: print("***", v, "doesnt match", line[k])
                    return False
        return True

    if not args.nopull or not os.path.isfile("%s/%s/dump.json" % (args.progdir, ".za.araali")):
        shutil.rmtree("%s/%s" % (args.progdir, ".za.araali"), ignore_errors=True)
        os.makedirs("%s/%s" % (args.progdir, ".za.araali"), exist_ok=True)
        with open("%s/%s/dump.json" % (args.progdir, ".za.araali"), "w") as f:
            zones_apps, status = API().get_zones_apps(full=True, tenant=args.tenant)
            if status == 0:
                # [{"zone_name": xx, "apps": [{"app_name": xx, "links": []}]}]
                for za in zones_apps:
                    json.dump(za, f)
                    f.write("\n")

    # Three types of searches: asset, workload, service, link (in future file and fork)
    if args.filters:
        filters = dict([make_map(a) for a in args.filters.split(",")])
    else:
        filters = {}
    for k in ["zone", "app"]:
        if k not in filters:
            filters[k] = ".*"

    search_type = "asset" # container, app/VM
    filter_keys = set(filters.keys())
    if set(["process", "binary_name", "parent_process"]) & filter_keys:
        search_type = "workload" # process lens
        if 0:
            for k in ["process", "binary_name", "parent_process"]:
                if k not in filters:
                    filters[k] = ".*"
    elif set(["dns_pattern", "dst_port"]) & filter_keys:
        search_type = "service" # service lens
        if 0:
            for k in ["dns_pattern", "dst_port"]:
                if k not in filters:
                    filters[k] = ".*"

    if args.verbose: print("filter:", filters)
    if args.verbose: print("search_type:", search_type)
    i = 0
    workload_dict = {}
    with open("%s/%s/dump.json" % (args.progdir, ".za.araali"), "r") as f:
        for line in f:
            zapps = json.loads(line)
            zname, apps = zapps["zone_name"], zapps["apps"]
            if not re.search(filters["zone"], zname):
                continue
            for app in apps:
                aname, links = app["app_name"], app.get("links", [])
                if not re.search(filters["app"], aname):
                    continue

                if 0 and search_type == "asset":
                    wkey = ":".join([zname, aname])
                    if wkey not in workload_dict:
                        print("="*40, i, "="*40)
                        i += 1
                        print(yaml.dump({"zone": zname, "app": aname}))
                        workload_dict[wkey] = 1
                    else:
                        workload_dict[wkey] += 1
                        raise # not expected to come here
                    continue

                for link in links:
                    # ['client', 'server', 'type', 'policy_skip_reason', 'error_code',
                    #  'state', 'timestamp', 'unique_id', 'alert_info', 'rollup_ids', 'active_ports']
                    client = link["client"]
                    server = link["server"]
                    # ['zone', 'app', 'unmapped_app', 'process', 'binary_name', 'parent_process']
                    if search_type == "asset":
                        nfilters = dict([a for a in filters.items()])

                        capp = client.get("app", "").split(".")
                        if len(capp) == 3:
                            client["app"] = capp[0]
                            client["pod"] = capp[1]
                            client["container"] = capp[2]

                        sapp = server.get("app", "").split(".")
                        if len(sapp) == 3:
                            server["app"] = sapp[0]
                            server["pod"] = sapp[1]
                            server["container"] = sapp[2]

                        client_m = match_filters(nfilters, client)
                        server_m = match_filters(nfilters, server)
                        if client_m:
                            wkey = [str(client[k]) for k in filters.keys()]
                            wkey = ":".join(wkey)
                            if wkey not in workload_dict:
                                print("="*40, i, "="*40)
                                i += 1
                                print(yaml.dump(client))
                                workload_dict[wkey] = 1
                            else:
                                workload_dict[wkey] += 1
                        
                        if server_m:
                            wkey = [str(server[k]) for k in filters.keys()]
                            wkey = ":".join(wkey)
                            if wkey not in workload_dict:
                                print("="*40, i, "="*40)
                                i += 1
                                print(yaml.dump(server))
                                workload_dict[wkey] = 1
                            else:
                                workload_dict[wkey] += 1

                    elif search_type == "workload":
                        nfilters = dict([a for a in filters.items() if a[0] not in ["zone", "app", "pod", "container"]])
                        client_m = match_filters(nfilters, client)
                        # {'subnet': '0.0.0.0', 'private_subnet': True} HOME
                        server_m = match_filters(nfilters, server)
                        # {'dns_pattern': '.*:dynamodb.*amazonaws.com:.*', 'dst_port': 443}

                        if search_type == "link":
                            if client_m or server_m:
                                print("="*40, i, "="*40)
                                i += 1
                                print(yaml.dump({"client": client, "server": server}))
                        else:
                            if client_m:
                                wkey = [str(a) for a in client.values()]
                                wkey.sort()
                                wkey = ":".join(wkey)
                                if wkey not in workload_dict:
                                    print("="*40, i, "="*40)
                                    i += 1
                                    print(yaml.dump(client))
                                    workload_dict[wkey] = 1
                                else:
                                    workload_dict[wkey] += 1

                            if server_m:
                                wkey = [str(a) for a in server.values()]
                                wkey.sort()
                                wkey = ":".join(wkey)
                                if wkey not in workload_dict:
                                    print("="*40, i, "="*40)
                                    i += 1
                                    print(yaml.dump(server))
                                    workload_dict[wkey] = 1
                                else:
                                    workload_dict[wkey] += 1
                    elif search_type == "service" and "dst_port" in server:
                        if "dns_pattern" in server:
                            dns_m = re.search(filters["dns_pattern"], server["dns_pattern"])
                            wkey = ":".join([server["dns_pattern"], str(server["dst_port"])])
                        elif "subnet" in server:
                            # {'subnet': '169.254.169.254', 'netmask': 32, 'dst_port': 80}
                            dns_m = re.search(filters["dns_pattern"], server["subnet"])
                            wkey = ":".join([server["subnet"], str(server["dst_port"])])
                        else:
                            dns_m = None
                            assert(server)
                        port_m = re.search(filters["dst_port"], str(server["dst_port"]))
                        if dns_m and port_m:
                            if wkey not in workload_dict:
                                print("="*40, i, "="*40)
                                i += 1
                                print(yaml.dump(server))
                                workload_dict[wkey] = 1
                            else:
                                workload_dict[wkey] += 1
        if 1:
            keys = list(workload_dict.keys())
            keys.sort(key=lambda x: -workload_dict[x])
            print("Found %s items" % len(keys))
            count = 0
            for i, k in enumerate(keys):
                print("%3s: %05s %s" % (i, workload_dict[k], k))
                count += workload_dict[k]
            print("Total count: %s" % count)

def helm(args):
    if args.nanny:
        print(api.API().get_helm_values(workload_name=args.zone,
                                        tenant=args.tenant,
                                        nanny=args.nanny))
    else:
        print(yaml.dump(araalictl.API().get_helm_values(
            args.zone, args.tenant)[0]))

def insights(args):
    insights, status = API().get_insights(args.zone, tenant=args.tenant)
    if status == 0:
        utils.dump_table(insights)

def token(args):
    if args.add:
        op = "add"
        name = args.add
    elif args.delete:
        op = "delete"
        name = args.delete
    elif args.list:
        op = "show"
        name = None
    tokens, status = araalictl.API().token(op=op, name=name, email=args.email, tenant=args.tenant)
    if status == 0:
        if op == "show":
            utils.dump_table(tokens["tokens"])
        else:
            print(tokens)

def ctl(args, remaining):
    api = araalictl.API()
    if remaining and remaining[0] not in ["authorize", "config"]:
        api.check()

    if remaining and remaining[0] == "--": remaining = remaining[1:]

    if remaining and remaining[0] in ["authorize", "upgrade"]:
        prepend = ["sudo"]
    else:
        prepend = []

    #p = subprocess.Popen(["sudo", cmdline, "authorize", "-token=-"], stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.STDOUT)
    #auth_stdout = p.communicate(input=api.cfg["token"].encode())[0]
    #print("echo %s | sudo %s authorize -token=-" % (api.cfg["token"], cmdline))
    #print(auth_stdout.decode())

    rc = os.system(" ".join(prepend + [api.cmdline] + remaining))
    return rc.returncode

def template(args):
    if args.t_subparser_name:
        return getattr(module_template, args.t_subparser_name)(args)

def fw_config(args):
    if not args.zone or not args.tenant:
        print("please specify zone and tenant")
        return

    if args.get:
        knobs = api.API().get_fw_config(zone=args.zone, tenant=args.tenant)
        print(knobs)
    elif args.update:
        status = api.API().update_fw_config(zone=args.zone,
                                            tenant=args.tenant,
                                            data_file_location=args.update)
        print(status)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description = 'Araali Python CLI')
    parser.add_argument('--verbose', '-v', action='count', default=0, help="specify multiple times to increase verbosity (-vvv)")
    parser.add_argument('--use_api', '-u', action='store_true', help="user api instead of araalictl")
    subparsers = parser.add_subparsers(dest="subparser_name")

    parser_config = subparsers.add_parser("config", help="list/change config params")
    parser_config.add_argument('-t', '--tenant', help="setup sub-tenant param (sticky)")
    parser_config.add_argument('--tenants', help="setup a list of sub-tenants (for managing alerts)")
    parser_config.add_argument('--backend', help="setup a custom backend (internal unit testing)")
    parser_config.add_argument('-g', '--get_tenants', action="store_true", help="get active tenants")
    parser_config.add_argument('-d', '--template_dir', help="custom template directory")

    parser_alerts = subparsers.add_parser("alerts", help="get alerts (to create templates for)")
    parser_alerts.add_argument('-t', '--tenant', help="get alert for a specific tenant")
    parser_alerts.add_argument('-n', '--nopull', action="store_true", help="dont pull from araali")
    parser_alerts.add_argument('-c', '--count', type=int, help="dont pull from araali")
    parser_alerts.add_argument('--ago', help="lookback")

    parser_insights = subparsers.add_parser("insights", help="get insights")
    parser_insights.add_argument('-t', '--tenant', help="get insights for a specific tenant")
    parser_insights.add_argument('-z', '--zone', help="insights for a specific zone")

    parser_assets = subparsers.add_parser("assets", help="get assets")
    parser_assets.add_argument('-t', '--tenant', help="get assets for a specific tenant")
    parser_assets.add_argument('-z', '--zone', help="assets for a specific zone")
    parser_assets.add_argument('-a', '--app', help="links for a specific app")
    parser_assets.add_argument('-n', '--nopull', action="store_true", help="dont pull from araali")
    parser_assets.add_argument('-f', '--filters', help="query filters. eg: zone=.*,app=.*,cve_sev=4,cve_count=^0$")
    parser_assets.add_argument('--ago', help="lookback")

    parser_token = subparsers.add_parser("token", help="manage api tokens")
    parser_token.add_argument('-t', '--tenant', help="token management for a specific tenant")
    parser_token.add_argument('-a', '--add', help="create new token")
    parser_token.add_argument('-d', '--delete', help="delete token by name")
    parser_token.add_argument('-e', '--email', help="user email")
    parser_token.add_argument('-l', '--list', action="store_true", help="list all tokens")

    parser_links = subparsers.add_parser("links", help="get links")
    parser_links.add_argument('-t', '--tenant', help="get links for a specific tenant")
    parser_links.add_argument('-z', '--zone', help="links for a specific zone")
    parser_links.add_argument('-a', '--app', help="links for a specific app")
    parser_links.add_argument('-s', '--svc', help="links for a specific svc")
    parser_links.add_argument('--ago', help="lookback")

    parser_za = subparsers.add_parser("search", help="search assets, workloads, services")
    parser_za.add_argument('-t', '--tenant', help="get zones and apps for a specific tenant")
    parser_za.add_argument('-n', '--nopull', action="store_true", help="dont pull from araali")
    parser_za.add_argument('-f', '--filters', help="query filters. eg: zone=.*,app=.*,pod=.*,container=.* process=.*,binary_name=.*,parent_process=.*,dns_pattern=.*,dst_port=.*")

    parser_helm = subparsers.add_parser("helm", help="generate values for araaly firewall helm chart")
    parser_helm.add_argument('-t', '--tenant', help="helm chart for a specific tenant")
    parser_helm.add_argument('-z', '--zone', help="helm for a specific zone")
    parser_helm.add_argument('-n', '--nanny', action="store_true", help="helm chart for nanny or firewall")

    parser_fw_config = subparsers.add_parser(
        "fw_config", help="modify araali fw config on the backend")
    parser_fw_config.add_argument(
        '-g', '--get', action="store_true", help="get the current config")
    parser_fw_config.add_argument(
        '-u', '--update', help="update with json in specified file location")
    parser_fw_config.add_argument(
        '-t', '--tenant', help="fw config for a specific tenant")
    parser_fw_config.add_argument(
        '-z', '--zone', help="fw config for a specific zone")

    parser_ctl = subparsers.add_parser("ctl", help="run araalictl commands")

    parser_template = subparsers.add_parser("template", help='Manage Templates as Code (in Git)')
    parser.add_argument('-T', '--template', help="apply operation for a specific template")

    t_subparsers = parser_template.add_subparsers(dest="t_subparser_name")

    parser_alerts = t_subparsers.add_parser("alerts", help="get alerts (to create templates for)")
    parser_alerts.add_argument('-t', '--tenant', help="get alert for a specific tenant")
    parser_alerts.add_argument('-n', '--nopull', action="store_true", help="dont pull from araali")
    parser_alerts.add_argument('-a', '--ago', help="dont pull from araali")

    parser_drift = t_subparsers.add_parser("drift", help="get drift (from template as code)")
    parser_drift.add_argument('-t', '--tenant', help="get drift for a specific tenant")
    parser_drift.add_argument('-p', '--public', action="store_true", help="drift in public library")
    parser_drift.add_argument('-n', '--nopull', action="store_true", help="dont pull from araali")
    parser_drift.add_argument('-T', '--template', help="get drift for a specific template (name or path)")

    parser_list = t_subparsers.add_parser("ls", help="list templates")
    parser_list.add_argument('-p', '--public', action="store_true", help="list from public library")
    parser_list.add_argument('-t', '--tenant', help="list for a sub-tenant")
    parser_list.add_argument('-T', '--template', help="list a specific template (name or path)")

    parser_pull = t_subparsers.add_parser("pull", help="pull templates")
    parser_pull.add_argument('-p', '--public', action="store_true", help="pull from public library")
    parser_pull.add_argument('-T', '--template', help="pull a specific template (name or path)")
    parser_pull.add_argument('-t', '--tenant')

    parser_push = t_subparsers.add_parser("push", help="push templates")
    parser_push.add_argument('-p', '--public', action="store_true", help="push to public library")
    parser_push.add_argument('-T', '--template', help="push a specific template (name or path)")
    parser_push.add_argument('-t', '--tenant', help="push for a sub-tenant")

    parser_format = t_subparsers.add_parser("fmt", help="rename template node name/ format into a normalized form")
    parser_format.add_argument('template')

    args, remaining = parser.parse_known_args()
    args.progdir, args.prog = os.path.split(sys.argv[0])

    if args.verbose:
        api.g_debug = True
        araalictl.g_debug = True

    if args.use_api:
        araalictl.g_use_api = True

    if args.subparser_name == "ctl":
        sys.exit(ctl(args, remaining))
    elif remaining:
        raise Exception("*** unknown flags: %s" % remaining)

    if args.template: # template can be specified by name or file path
        args.dirname, args.pathname = os.path.split(args.template)
        args.template = args.pathname.split(".yaml")[0]

    if args.subparser_name:
        sys.exit(locals()[args.subparser_name](args))
    else:
        araalictl.API().check()

    sys.exit(0)



"""fake api using araalictl underneath
    api = api.API()

    alerts, page, status = api.get_alerts(args.count, args.ago)
    assets, status = api.get_assets(args.zone, args.app, args.ago)
    links, status = api.get_links(args.zone, args.app, args.svc, args.ago)
    insights, status = api.get_insights(args.zone)
"""
from __future__ import print_function

from . import api
import datetime
import os
import platform
import requests
import subprocess
import sys
from . import utils
import yaml

g_debug = False
g_use_api = False

class API:
    def __init__(self):
        if g_use_api:
            self.api = api.API()
            return

        def fetch():
            """For downloading and upgrading araalictl"""
            progdir = os.path.dirname(os.path.abspath(__file__))
            afile = progdir + "/bin/araalictl"
            rc = utils.run_command("mkdir -p %s/bin" % (progdir), debug=g_debug, result=True, strip=False)
            afile = progdir + "/bin/araalictl"
            if os.path.isfile(afile):
                #print("upgrading araalictl ...")
                #rc = utils.run_command("sudo %s upgrade" % (afile), debug=g_debug, result=True, strip=False)
                return afile

            url = {
                    "Linux": "https://s3-us-west-2.amazonaws.com/araalinetworks.cf/araalictl-api.linux-amd64",
                    "Darwin": "https://s3-us-west-2.amazonaws.com/araalinetworks.cf/araalictl-api.darwin-amd64"
                  }.get(platform.system(), None)
            if url is None:
                raise Exception("*** only linux and mac are currently supported: %s" % platform)

            print("Fetching araalictl ...")
            r = requests.get(url, allow_redirects=True)
            open(afile, 'wb').write(r.content)
            os.chmod(afile, 0o777)
            return afile

        self.cmdline = fetch()

    def check(self):
        # nightly.aws.araalinetworks.com
        rc = utils.run_command("%s config InternalCfgBackend" % (self.cmdline),
                                debug=g_debug, result=True, strip=False)
        try:
            backend = rc[1].decode().split("=", 1)[1].split(".")[0]
        except:
            backend = None

        if backend != utils.cfg["backend"]:
            print("*** backend doesnt match, invoking deauth ...")
            os.system(("sudo %s authorize -clean" % (self.cmdline)))

        rc = utils.run_command("%s authorize check" % (self.cmdline),
                                debug=g_debug, result=True, strip=False)
        ret = rc[1].decode().strip()
        if ret != "copy is authorized":
            print("*** not authorized for %s, must re-authorize in Araali UI" % utils.cfg["backend"])
            rc = utils.run_command("sudo %s authorize -clean" % (self.cmdline),
                                debug=g_debug, result=True, strip=False)
            rc = utils.run_command("sudo rm -rf /tmp/actl* /tmp/av.port",
                                debug=g_debug, result=True, strip=False)
            cmdline = "sudo %s authorize -bend %s" % (self.cmdline, utils.cfg["backend"])
            #cmdline = "sudo %s authorize -local" % (self.cmdline)
            os.system(cmdline)

        if "email" not in utils.cfg:
            # DevName=blah@blah.com
            rc = utils.run_command("%s config DevName" % (self.cmdline),
                                debug=g_debug, result=True, strip=False)
            utils.cfg["email"] = rc[1].decode().split("=", 1)[1]

    def get_tenants(self):
        self.check()

        rc = utils.run_command("%s api -fetch-active-subtenants" % (
                                self.cmdline),
                                debug=g_debug, result=True, strip=False)

        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def get_alerts(self, count=None, ago=None, token=None, tenant=None):
        """get alerts"""

        if g_use_api:
            return self.api.get_alerts(count, ago, token, tenant)

        self.check()
        if tenant is None: tenant = utils.cfg["tenant"]
        if count is None: count = 1000

        if not ago:
            start_time = 0
            end_time = 0
        else:
            ago = dict([utils.make_map(a) for a in ago.split(",")])
            start_time = datetime.datetime.now() - datetime.timedelta(**ago)
            end_time = datetime.datetime.now()

            start_time = int(start_time.timestamp())
            end_time = int(end_time.timestamp())

        cmd = "-starttime %s" % (start_time)
        cmd += " -endtime %s" % (end_time)
        if token:
            cmd += " -paging-token %s" % token

        tstr = " -tenant=%s " % (tenant) if tenant else ""
        rc = utils.run_command("%s api -fetch-alerts %s -count %s %s" % (
                                self.cmdline, cmd, count, tstr),
                                debug=g_debug, result=True, strip=False)

        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        token = None
        objs = yaml.load(rc[1], yaml.SafeLoader)
        if objs:
            token=objs[-1].get("paging_token", None)

        return objs, token, rc[0]

    def get_assets(self, zone=None, app=None, ago=None, tenant=None):
        """Get assets for a zone (required) and app (optional)
            Usage: assets, status = api.get_assets()
        """

        if g_use_api:
            return self.api.get_assets(zone, app, ago, tenant)

        self.check()
        if not ago:
            ago = "days=1"
        ago = dict([utils.make_map(a) for a in ago.split(",")])
        start_time = datetime.datetime.now() - datetime.timedelta(**ago)
        end_time = datetime.datetime.now()

        cmd = "-starttime %s" % (int(start_time.timestamp()))
        cmd += " -endtime %s" % (int(end_time.timestamp()))
        if zone: cmd += " -zone %s" % zone
        if app: cmd += " -app %s" % app
        if tenant is None: tenant = utils.cfg["tenant"]
        if tenant: cmd += " -tenant %s" % tenant

        rc = utils.run_command("%s api %s -fetch-compute" % (
            self.cmdline, cmd), debug=g_debug, result=True, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def get_links(self, zone=None, app=None, svc=None, ago=None, tenant=None):
        """Get links for a zone and app, or svc"""

        if g_use_api:
            return self.api.get_links(zone, app, svc, ago, tenant)
        self.check()
        assert svc is not None or zone is not None and app is not None

        if tenant is None: tenant = utils.cfg["tenant"]
        tstr = " -tenant=%s " % (tenant) if tenant else ""
        if svc:
            tstr += " -service=%s " % (svc)
        else:
            tstr += " -zone=%s -app=%s " % (zone, app)

        if not ago:
            ago = "days=1"
        ago = dict([utils.make_map(a) for a in ago.split(",")])
        start_time = datetime.datetime.now() - datetime.timedelta(**ago)
        end_time = datetime.datetime.now()
        tstr += " -starttime %s" % (int(start_time.timestamp()))
        tstr += " -endtime %s" % (int(end_time.timestamp()))

        rc = utils.run_command("%s api -fetch-links %s" % (self.cmdline, tstr),
                         debug=g_debug, result=True, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def get_insights(self, zone=None, tenant=None):
        """Get insights for a zone (optional)"""

        if g_use_api:
            return self.api.get_insights(zone, tenant)
        self.check()
        cmd = ""
        if zone: cmd += " -zone %s" % zone
        if tenant is None: tenant = utils.cfg["tenant"]
        if tenant: cmd += " -tenant %s" % tenant

        rc = utils.run_command("%s api %s -fetch-insights" % (
            self.cmdline, cmd), debug=g_debug, result=True, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def token(self, op=None, name=None, email=None, tenant=None):
        """token management"""

        self.check()
        if op == None:
            print("*** operation not specified")
            return
        if email is None: email = utils.cfg["email"]
        if op == "add":
            cmd = "-name=%s %s" % (name, email)
        elif op == "delete":
            cmd = "-delete -name=%s %s" % (name, email)
        else:
            cmd = "-list"

        #if tenant is None: tenant = utils.cfg["tenant"]
        #if tenant: cmd += " -tenant=%s" % tenant

        rc = utils.run_command("%s token -api-access %s" % (
            self.cmdline, cmd), debug=g_debug, result=True, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def get_templates(self, public=None, template=None, tenant=None):
        """Get templates (or public ones)"""

        self.check()

        if tenant is None: tenant = utils.cfg["tenant"]
        tstr = " -tenant=%s " % (tenant) if tenant else ""
        tstr += " -public " if public else ""
        tstr += " -template=%s " % (template) if template else ""

        rc = utils.run_command("%s api -list-templates %s" % (
            self.cmdline, tstr), debug=g_debug, result=True, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def update_template(self, data, public=False, tenant=None):
        """Update template"""

        self.check()

        if tenant is None: tenant = utils.cfg["tenant"]
        tstr = " -tenant=%s " % (tenant) if tenant else ""

        command = "-update-template" if not public else "-export-template"
        rc = utils.run_command("%s api %s %s" % (
            self.cmdline, command, tstr), in_text=yaml.dump(data), debug=g_debug, result=True, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def get_zones_apps(self, full=False, tenant=None):
        """Get zones and apps"""
        self.check()

        if tenant is None: tenant = utils.cfg["tenant"]
        tstr = " -tenant=%s " % (tenant) if tenant else ""
        rc = utils.run_command("%s api -fetch-zone-apps %s %s" % (self.cmdline, "-full" if full else "", tstr),
                result=True, debug=g_debug, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def get_helm_values(self, zone=None, tenant=None):
        """Get helm chart"""
        self.check()

        if tenant is None: tenant = utils.cfg["tenant"]
        tstr = " -tenant=%s " % (tenant) if tenant else ""
        if zone is not None:
            tstr += " -tags=zone=%s " % zone
        rc = utils.run_command("%s fortify-k8s -out helm %s" % (self.cmdline, tstr),
                result=True, debug=g_debug, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]

    def rbac_show_users(self, tenant=None):
        """show rbac"""

        self.check()
                                                     
        tstr = " -tenant=%s " % (tenant) if tenant else ""
        rc = utils.run_command("%s user-role -op list-user-roles %s" % (
                         self.cmdline, tstr), result=True, debug=g_debug, strip=False)
        if rc[0] != 0:
            print("*** failed: %s" % rc[1].decode())
        return yaml.load(rc[1], yaml.SafeLoader), rc[0]




#!/usr/bin/env python

import sys

count_dict = {}
line_type = "unique"
for line in sys.stdin:
    line = line.strip()
    if "# unique" in line:
        continue
    if not line:
        line_type = "new"
        continue
    if line_type == "unique":
        count_dict[line] = 1
        continue
    if line_type == "new":
        line_type = line
        count_dict[line] = 0
    count_dict[line_type] += 1

keys = list(count_dict.keys())
keys.sort(key=lambda x: -count_dict[x])
for k in keys:
    print("%-70s %s" % (k, count_dict[k]))
    
